#include <stdio.h>
#include <stdlib.h>
#include <signal.h>
#include <sys/ipc.h>
#include <sys/shm.h>
#include <sys/sem.h>
#include <sys/types.h>
#include <string.h>
#include <errno.h>
#include "common.h"

union sema_undef
{
	struct semid_ds *buf;
	ushort *array;
	int val;
};

// 공유메모리 생성
void CreateSharedMemory();

// 세마포어 생성
void CreateSemaphore();

// 세마포어 삭제
void DeleteSemaphore();

// 공유메모리 삭제
void DeleteSharedMemory();

// 시그널 등록
void InitializeSignal();

// 종료 시그널 처리 핸들러
void SIGINT_Handler(int signo);

// 유저 접속한 경우
void SIGUSR1_Handler(int signo);

// 유저 접속해제 한경우
void SIGUSR2_Handler(int signo);

int shared_memory_id = 0;
int semaphore_id = 0;
void *pShemBuffer = NULL;

// SIG_INT OR SIG_TERM 받은경우 TRUE로 변환
int isReceivedSIGINT = FALSE;

// 종료
int isExitStatus = FALSE;

int main(void)
{
	InitializeSignal();
	
	CreateSharedMemory(&shared_memory_id, &pShemBuffer);

	CreateSemaphore();
	
	// SharedData 구조체에 서버 데이터 저장 후 메모리 복사 진행	
	SharedData sharedData;

	memset((void*)&sharedData, 0, sizeof(sharedData));

	sharedData.Server.pid = getpid();
	strcpy(sharedData.Server.name, SERVER_STRING);

	memcpy(pShemBuffer, (void*)&sharedData, sizeof(SharedData));

	printf("Chatting server started!\n");

	// 종료될때까지 반복
	while(!isExitStatus);

	DeleteSharedMemory();

	DeleteSemaphore();

	return 0;
}

// 공유메모리 생성
void CreateSharedMemory()
{
	shared_memory_id = shmget(SHARED_MEMORY_KEY, SHARED_MEMORY_SIZE, 0644 | IPC_CREAT | IPC_EXCL);

	if(-1 == shared_memory_id)
	{	
		if(EEXIST == errno)
		{
			printf("server exists\n");	
		}
		else
		{
			printf("shmget error\n");
		}

		exit(1);
	}
	
	pShemBuffer = shmat(shared_memory_id, NULL, 0);
	
	if((void*)-1 == pShemBuffer)
	{
		printf("Shared Memory Buffer Create Failed...\n");
		exit(1);
	}	
}

// 종료 시그널 처리
void SIGINT_Handler(int signo)
{
	int i = 0;
	SharedData *pSharedData = (SharedData*)pShemBuffer;

	if(NULL == pSharedData)
	{
		printf("SharedData is NULL\n");
		exit(1);
	}

	if(pSharedData->numOfProcess > 0)
	{
		for(i = pSharedData->numOfProcess -1; i >= 0; i--)
		{
			kill(pSharedData->pData[i].pid, SIGINT);
			sleep(1);
		}
	}	

	isReceivedSIGINT = TRUE;

	if(pSharedData->numOfProcess == 0)
	{
		isExitStatus = TRUE;
	}
}

// 유저가 접속한 경우
void SIGUSR1_Handler(int signo)
{
	SharedData *pSharedData = (SharedData*)pShemBuffer;
	int index = 0;
	int i = 0;

	if(NULL == pSharedData)
	{
		printf("SharedData is NULL\n");
		exit(1);
	}

	index = pSharedData->numOfProcess - 1;
	
	printf("# [%s] has entered. #\n\n", pSharedData->pData[index].name);

	if(pSharedData->numOfProcess == 2)
	{	
		for(i = 0; i < pSharedData->numOfProcess; i++)
		{
			kill(pSharedData->pData[i].pid, SIGUSR1);
		}

		printf("# [%s]와 [%s]가 채팅을 즐기고 있습니다.\n\n",pSharedData->pData[0].name, pSharedData->pData[1].name);
	}
}

// 유저가 접속 해제한 경우
void SIGUSR2_Handler(int signo)
{
	SharedData *pSharedData = (SharedData*)pShemBuffer;
	int i = 0;

	if(NULL == pSharedData)
	{
		printf("SharedData is NULL\n");
		return;
	}

	printf("# [%s] disconnect #\n\n",pSharedData->pData[pSharedData->numOfProcess].name);

	// 남아있는 유저에게 상대방 접속해제 신호 전달
	for(i = 0; i < pSharedData->numOfProcess; i++)
	{
		kill(pSharedData->pData[i].pid, SIGALRM);
	}

	// 종료하라는 메시지를 받고 프로세스 수가 0이라면 종료
	if(isReceivedSIGINT && pSharedData->numOfProcess == 0)
	{
		isExitStatus = TRUE;
	}
}
 
void DeleteSharedMemory()
{
	if(-1 == shmdt(pShemBuffer))
	{
		printf("shmdt error\n");
	}

	if(-1 == shmctl(shared_memory_id, IPC_RMID, 0))
	{
		printf("shmctl error\n");
		exit(1);
	}
}		

// 시그널 등록
void InitializeSignal()
{
	if(SIG_ERR == signal(SIGTERM, SIGINT_Handler))
	{
		printf("cannot handler SIGINT\n");
		exit(EXIT_FAILURE);
	}

	if(SIG_ERR == signal(SIGINT, SIGINT_Handler))
	{
		printf("cannot handler SIGINT\n");
		exit(EXIT_FAILURE);
	}

	if(SIG_ERR == signal(SIGUSR1, SIGUSR1_Handler))
	{
		printf("cannot handler SIGUSR1\n");
		exit(EXIT_FAILURE);
	}

	if(SIG_ERR == signal(SIGUSR2, SIGUSR2_Handler))
	{
		printf("cannot handler SIGUSR2\n");
		exit(EXIT_FAILURE);
	}
}

// 세마포어 생성
void CreateSemaphore()
{
	union sema_undef sema;		

	sema.val = 1;

	semaphore_id = semget(SEMAPHORE_KEY, 1, IPC_CREAT | IPC_EXCL | 0666);

	if(-1 == semaphore_id)
	{
		printf("semget error\n");

		DeleteSharedMemory();
		
		exit(1);
	}

	if(-1 == semctl(semaphore_id, 0, SETVAL,sema))
	{
		DeleteSharedMemory();

		printf("cannot set semaphore value.\n");
		exit(1);
	}
}

// 세마포어 삭제
void DeleteSemaphore()
{
	if(-1 == semctl(semaphore_id, 0, IPC_RMID))
	{
		printf("sempahore delete fail\n");
		exit(1);
	}
}
